from useis.services.grid_service import client
from uquake.grid.nlloc import VelocityGrid3D
from uquake.nlloc.nlloc import Srces
from uquake.core import read_inventory
from importlib import reload
reload(client)

root_dir = 'projects'
test_project = 'test_project'
test_network = 'TN'

grid_dim = [10, 10, 10]
grid_spacing = [1, 1, 1]
grid_origin = [0, 0, 0]


def create_velocity():
    p_velocity_grid = VelocityGrid3D(test_network,
                                     grid_dim,
                                     grid_origin,
                                     grid_spacing,
                                     phase='P', value=5000)
    s_velocity_grid = VelocityGrid3D(test_network,
                                     grid_dim,
                                     grid_origin,
                                     grid_spacing,
                                     phase='S', value=3000)

    return p_velocity_grid, s_velocity_grid


p_velocity, s_velocity = create_velocity()
gc = client.GridClient('http://127.0.0.1:8000/', test_project, test_network)

response = gc.add_3d_velocity(p_velocity)
response = gc.add_3d_velocity(s_velocity)
# inv = read_inventory('/home/jpmercier/Repositories/Nanometrics/data'
#                      '/OSA_Light.xml', xy_from_lat_lon=True)
# response = gc.add_inventory(inv)
srces = Srces.generate_random_srces_in_grid(p_velocity,
                                            n_srces=20)
response = gc.add_srces(srces)
response = gc.initialize_travel_times(multi_threaded=True)
# print(gc.list_3d_velocity())

response, points = gc.generate_random_points_in_grid(n_points=1)

x = points[0][0]
y = points[0][1]
z = points[0][2]

response, observations = gc.generate_random_observations(x, y, z)

response = gc.event_location(observations)



import pandas as pd

from ..core.project_manager import ProjectManager
from pathlib import Path
import os
import shutil
from ..settings.settings import Settings
from uquake.waveform.pick import snr_ensemble_re_picker, calculate_snr
from uquake.core.event import (Pick, WaveformStreamID, ResourceIdentifier,
                               Arrival, Origin, Event)
from uquake.core import UTCDateTime
import uquake
from ..ai.model import AIPicker
from datetime import datetime
from uquake.core.stream import Stream
from uquake.core.inventory import Inventory


class PickerResult(object):
    def __init__(self, new_picks, initial_picks, snr_threshold, stream):
        self.new_picks = new_picks
        self.initial_picks = initial_picks
        self.snr_threshold = snr_threshold
        self.stream = stream

        for k, pick in enumerate(self.new_picks):
            network = pick.waveform_id.network_code
            station = pick.waveform_id.station_code
            location = pick.waveform_id.location_code

            snr = calculate_snr(self.stream.select(network=network,
                                                   station=station,
                                                   location=location
                                                   )[0],
                                pick.time,
                                pre_wl=0.05, post_wl=0.02)

            self.new_picks[k].snr = snr

    def export_event(self):
        return self.append_event()

    @property
    def picks(self):
        picks = []
        for pick in self.new_picks:
            if pick.snr < self.snr_threshold:
                continue
            picks.append(pick)

        return picks

    def append_event(self, event: uquake.core.event.Event = None):

        snr_threshold = self.snr_threshold
        arrivals = []
        picks = []

        for pick in self.new_picks:
            picks.append(pick)

        for pick in self.initial_picks:
            picks.append(pick)

            arrival = Arrival(pick_id=pick.resource_id,
                              phase=pick.phase_hint)
            arrivals.append(arrival)

            origin = Origin(arrivals=arrivals)

        if event is None:
            event = Event(picks=picks, origins=[origin])

            return event

        if isinstance(event, uquake.core.event.Catalog):
            event[0].origins.append(origin)
            for pick in self.new_picks:
                event[0].picks.append(pick)

            return event

        else:
            event.origins.append(origin)
            for pick in self.new_picks:
                event.picks.append(pick)

            return event

    # def measure_incidence(self, stream: Stream,
    #                       inventory: Inventory):
    #
    #     pass

    def write_simple_pick_file(self, filename):
        with open(filename, 'w') as pick_file_out:
            for pick in self.picks:
                network = pick.waveform_id.network_code
                station = pick.waveform_id.station_code
                location = pick.waveform_id.location_code
                channel = None
                if pick.waveform_id.channel_code is not None:
                    channel = pick.waveform_id.channel_code[0:2]

                if pick.snr < self.snr_threshold:
                    continue

                station_string = f'{network}.{station}.{location}.{channel}*'
                line_out = f'{station_string},{pick.phase_hint.upper()},' \
                           f'{pick.time.timestamp}\n'
                pick_file_out.write(line_out)


class Picker(ProjectManager):

    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str):
        """
        Object to manage repicking within a project
        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        """

        super().__init__(base_projects_path, project_name, network_code)

        self.files.picker_settings = self.paths.config / 'picker_settings.toml'

        if not self.files.picker_settings.is_file():
            settings_template = Path(os.path.realpath(__file__)).parent / \
                            '../settings/picker_settings_template.toml'

            shutil.copyfile(settings_template,
                            self.files.picker_settings)

        self.settings = Settings(settings_location=self.paths.config,
                                 settings_file=
                                 self.files.picker_settings.name)

        self.files.ai_picker_model = self.paths.ai_models / \
                                     'picker_model.pickle'

        if self.files.ai_picker_model.is_file():
            if self.files.ai_picker_model.is_file():
                self.ai_picker = AIPicker.read(
                    self.files.ai_picker_model)

    @staticmethod
    def read_nmx_pick(pick_file: str):

        nmx_picks = pd.read_csv(pick_file,
                                names=['site_id', 'phase', 'timestamp'])

        nmx_picks['datetime'] = [datetime.utcfromtimestamp(
            pick[1]['timestamp']) for pick in nmx_picks.iterrows()]
        out_picks = []
        for pick in nmx_picks.iterrows():
            pick = pick[1]
            network, station, location, channel = pick.site_id.split('.')

            waveform_id = WaveformStreamID(network_code=network,
                                           station_code=station,
                                           location_code=location,
                                           channel_code=channel)

            out_picks.append(
                Pick(waveform_id=waveform_id, phase_hint=pick.phase,
                     time=UTCDateTime(pick.datetime),
                     evaluation_mode='automatic',
                     evaluation_status='preliminary',
                     method_id=ResourceIdentifier('beamforming'),
                     method='beamforming'))

        return out_picks

    def add_ai_picker_from_file(self, file_path):
        shutil.copyfile(file_path, self.files.ai_picker_model)
        self.ai_picker = AIPicker.read(self.files.ai_picker_model)

    def snr_repick(self, stream: uquake.core.stream.Stream,
               picks: uquake.core.event.Pick):

        """
        Repick using a snr based ensemble picker
        :param stream: stream containing the waveforms
        :type stream: :class: uquake.core.stream.Stream
        :param picks: a list of picks associated to the waveforms
        :type picks: a list of :class: uquake.core.event.Pick
        :return:
        """

        start_search_window = self.settings.snr_repicker.start_search_window
        end_search_window = self.settings.snr_repicker.end_search_window
        start_refined_search_window = \
            self.settings.snr_repicker.start_refined_search_window
        end_refined_search_window = \
            self.settings.snr_repicker.end_refined_search_window
        search_resolution = self.settings.snr_repicker.search_resolution
        pre_pick_window_len = self.settings.snr_repicker.pre_pick_window_len
        post_pick_window_len = self.settings.snr_repicker.post_pick_window_len

        snrs, new_picks = snr_ensemble_re_picker(stream, picks,
                                                 start_search_window,
                                                 end_search_window,
                                                 start_refined_search_window,
                                                 end_refined_search_window,
                                                 search_resolution,
                                                 snr_calc_pre_pick_window_len
                                                 =pre_pick_window_len,
                                                 snr_calc_post_pick_window_len
                                                 =post_pick_window_len)

        out_picks = []
        for _i, new_pick in enumerate(new_picks):
            network = new_pick.waveform_id.network_code
            station = new_pick.waveform_id.station_code
            location = new_pick.waveform_id.location_code
            channel = new_pick.waveform_id.channel_code

            pre_wl = self.settings.snr_repicker.snr_pre_pick_window
            post_wl = self.settings.snr_repicker.snr_post_pick_window

            snr = calculate_snr(stream.select(network=network, station=station,
                                              location=location)[0],
                                new_pick.time, pre_wl=pre_wl, post_wl=post_wl)

            # if snr < self.settings.snr_repicker.snr_threshold:
            #     continue

            new_pick.snr = snr
            out_picks.append(new_pick)

        return PickerResult(new_picks=out_picks,
                            initial_picks=picks,
                            snr_threshold=
                            self.settings.snr_repicker.snr_threshold,
                            stream=stream)



from pathlib import Path
import pickle
from uquake.core.inventory import read_inventory
from uquake.core.logging import logger
from uquake.nlloc import (Srces)
from uquake.core.event import ConfidenceEllipsoid, OriginUncertainty
from uquake.grid import nlloc as nlloc_grid
from uquake.grid import read_grid
from time import time
import numpy as np
import os
import shutil
from ..settings.settings import Settings
from uquake.core.event import AttribDict


def read_srces(fname):
    with open(fname, 'rb') as srces_file:
        return pickle.load(srces_file)


# def read_inventory():
#     if self.files.srces.exists() and use_srces:
#         with open(self.files.srces, 'rb') as srces_file:
#             logger.info('srces will be read from the file and not build'
#                         'from the inventory file')
#             self.srces = pickle.load(srces_file)
#
#         if self.files.inventory.exists():
#             self.inventory = read_inventory(str(self.files.inventory))
#
#     elif self.files.inventory.exists():
#         self.inventory = read_inventory(str(self.files.inventory))
#         self.srces = Srces.from_inventory(self.inventory)
#         logger.info('srces will be build from the inventory file. The '
#                     'srces.pickle file will be replaced.')
#         with open(self.files.srces, 'wb') as srces_file:
#             pickle.dump(self.srces, srces_file)
#
#     elif self.files.srces.exists():
#         logger.info('no inventory file in the project. srces file will be '
#                     'used instead.')
#         with open(self.files.srces, 'rb') as srces_file:
#             self.srces = pickle.load(srces_file)


class ProjectManager(object):

    inventory_file_name = 'inventory.xml'

    def __init__(self, base_projects_path, project_name, network_code,
                 use_srces=False, **kwargs):
        """
        Interface to manage project and grids.

        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        :param use_srces: if True use the srces files instead of the the
        inventory file should both files be present (default=False)
        :Example:
        >>> from uquake.grid import nlloc as nlloc_grid
        >>> from uquake.nlloc import nlloc
        # initialize the project with the root path of the project,
        the project and the network names
        >>> project_name = 'test'
        >>> network_code = 'test'
        >>> root_path=''
        >>> pm = nlloc.ProjectManager(root_path, project_name, network_code)
        # this will initialize the project and create a specific directory
        # structure if the directories do not exist.

        # add an inventory file to the project
        >>> path_to_uquake_inventory_file = 'PATH/TO/THE/INVENTORY/FILE.xml'

        ..note :: the uquake inventory object inherit from the obspy inventory
                  and
                  behaves in a very similar way. It, however, differs from
                  its parent has it implements properties specific to local
                  earthquake who are either expressed using UTM coordinates
                  system or a local coordinates system that cannot necessarily
                  be translated to latitude and longitude. Beyond this minor
                  difference, uquake inventory structure also differs from
                  the standard inventory and implements a slightly different
                  hierarchy more representative of mine monitoring systems.
                  The uquake Inventory object hierarchy is as follows:
                 1. Inventory
                    1.1. Networks: A useis network. a network includes at
                                   least one data acquisition station.
                        1.1.1 Stations: A place where the data acquisition is
                                        performed. For instance, station
                                        usually includes power, communication
                                        and data acquisition equipment. One
                                        or more site can be connected to a
                                        station.
                            1.1.1.1 sites: A instrument converting a
                                             physical phenomenon to data either
                                             digital or analog. A site
                                             comprises one or more channel.
                                1.1.1.1.1 Channels: A channel is a the smallest
                                                    unit of measuring.


        >>> inventory = read_inventory(path_to_uquake_inventory_file)
        >>> pm.add_inventory(inventory)

        # alternatively, sites can be added to the using the srces object.
        # sites can be added this way from a nlloc.nlloc.Srces object using
        the .add_srces method.
        ..note:: srces stands for sources and it is the nomenclature used in
                 NonLinLoc. This might be a soure of confusion for the users.
                 In addition, what NonLinLoc refers to as station is called a
                 sites in this context.

        # srces object can be constructed from an inventory as follows:
        >>> srces = Srces.from_inventory(inventory)
        # alternatively, each sites can be added individually using the
        .add_site method. As follows:
        >>> srces = Srces()
        >>> x = 250
        >>> y = 250
        >>> z = 250
        >>> elevation = 0
        >>> srces.add_site('site label', x, y, z, elev=elevation)

        # Srces can be added to the project as follows:
        >>> pm.add_srces(srces)

        # add the velocity models to the project. P- and S-wave velocity models
        can be added separately from a nlloc.grid.VelocityGrid3D
        using the .add_velocity method or from a nlloc.grid.VelocityEnsemble
        object using the .add_velocities method

        >>> origin = [0, 0, 0]
        >>> spacing = [25, 25, 25]
        >>> dimensions = [100, 100, 100]

        >>> nlloc_grid.VelocityGrid3D()
        >>> vp = 5000  # P-wave velocity in m/s
        >>> vs = 3500  # S-wave velocity in m/s
        >>> p_velocity = nlloc_grid.VelocityGrid3D(network_code, dimensions,
        >>>                                        origin, spacing, phase='P',
        >>>                                        value=5000)
        >>> s_velocity = nlloc_grid.VelocityGrid3D(network_code, dimensions,
        >>>                                        origin, spacing, phase='S',
        >>>                                        value=5000 )
        >>> pm.add_velocity(p_velocity)
        >>> pm.add_velocity(s_velocity)
        # Alternatively
        >>> velocities = nlloc_grid.VelocityGridEnsemble(p_velocity,
        >>>                                              s_velocity)
        >>> pm.add_velocities(velocities)
        # Adding a velocity model of the velocity models triggers the
        # calculation of the travel time grids.
        # It is possible to manually trigger the calculation of the travel
        # time grid by invoking
        >>> pm.init_travel_time_grids()
        # this should not, however, be required.

        # prior to running the location, NonLinLoc need to be configured.
        # configuring NonLinLoc can be done using the nlloc.nlloc.NonLinLoc
        # object
        >>> nonlinloc = nlloc.NonLinLoc()
        # this will initialize the nonlinloc object sith default value. Those
        # value have been used to locate useis events in a volumes of
        # approximately 3000 m x 3000 m x 1500 m. The parameters should be
        # provide adequate results for volumes of similar scale but would need
        # to be adapted to smaller or larger volumes.

        """

        self.project_name = project_name
        self.network_code = network_code

        p_vel_base_name = nlloc_grid.VelocityGrid3D.get_base_name(network_code,
                                                                  'P')
        s_vel_base_name = nlloc_grid.VelocityGrid3D.get_base_name(network_code,
                                                                  'S')

        self.base_projects_path = Path(base_projects_path)
        base_project_path = Path(base_projects_path)
        root = base_project_path / project_name / network_code
        self.paths = {'base': base_project_path,
                      'root': base_project_path / project_name / network_code,
                      'archives': base_project_path / 'archives',
                      'inventory': root / 'inventory',
                      'config': root / 'config',
                      'velocities': root / 'velocities',
                      'times': root / 'times',
                      'hdf5_times': root / 'hdf5_times',
                      'ai_models': root / 'ai_models'}

        self.paths = AttribDict(self.paths)

        self.files = {'inventory': self.paths.inventory / 'inventory.xml',
                      'srces': self.paths.inventory / 'srces.pickle',
                      'settings': self.paths.config / 'settings.toml',
                      'services_settings': self.paths.config /
                                           'services_settings',
                      'p_velocity': self.paths.velocities / p_vel_base_name,
                      's_velocity': self.paths.velocities / s_vel_base_name,
                      'hdf5_tt': self.paths.hdf5_times /
                                 'travel_times_tables.h5f',
                      'classification_model': self.paths.ai_models /
                                              'classification.pickle'}

        self.files = AttribDict(self.files)

        # create the directory if it does not exist
        self.paths.root.mkdir(parents=True, exist_ok=True)
        for key in self.paths.keys():
            path = self.paths[key]
            logger.info(path)
            path.mkdir(parents=True, exist_ok=True)

        # SETTINGS

        if not self.files.settings.is_file():
            settings_template = Path(os.path.realpath(__file__)).parent / \
                                    '../settings/settings_template.toml'

            shutil.copyfile(settings_template, self.files.settings)

        if not self.files.services_settings.is_file():
            settings_path = Path(os.path.realpath(__file__)).parent.parent
            settings_template = settings_path / \
                                'settings/services_settings_template.toml'

            shutil.copyfile(settings_template, self.files.services_settings)

        self.settings = Settings(str(self.paths.config))

        self.srces = None
        self.inventory = None

        if self.files.inventory.exists():
            self.inventory = read_inventory(str(self.files.inventory))
            self.srces = Srces.from_inventory(self.inventory)
        elif self.files.srces.exists():
            with open(self.files.srces, 'rb') as srces_file:
                self.srces = pickle.load(srces_file)

        self.p_velocity = None
        try:
            fname = str(self.paths.velocities / p_vel_base_name)
            self.p_velocity = read_grid(fname, format='NLLOC')
        except Exception as e:
            logger.warning(e)

        self.s_velocity = None
        try:
            fname = str(self.paths.velocities / s_vel_base_name)
            self.s_velocity = read_grid(fname, format='NLLOC')
        except Exception as e:
            logger.warning(e)

        self.velocities = None
        if (self.p_velocity is not None) and (self.s_velocity is not None):
            self.velocities = nlloc_grid.VelocityGridEnsemble(self.p_velocity,
                                                              self.s_velocity)

        self.paths.times = self.paths.root / 'times'
        self.paths.times.mkdir(parents=True, exist_ok=True)
        file_list = list(self.paths.times.glob('*time*'))

        self.travel_times = None
        try:
            self.travel_times = nlloc_grid.TravelTimeEnsemble.from_files(
                self.paths.times)
        except Exception as e:
            logger.error(e)

    @staticmethod
    def exists(project_path, project_name, network_code):
        project_path = Path(project_path) / project_name / network_code
        return project_path.exists()

    def init_travel_time_grids(self, multi_threaded=True):
        """
        initialize the travel time grids
        """

        has_velocity = self.has_p_velocity() or self.has_s_velocity()

        abort = not (has_velocity and self.has_inventory())

        if not self.has_p_velocity():
            logger.warning('No P-wave velocity found in project')

        if not self.has_s_velocity():
            logger.warning('No S-wave velocity found in project')

        if not self.has_inventory():
            logger.warning('No inventory file found in project')

        if abort:
            logger.warning('travel time grids will not be calculated')
            return

        logger.info('initializing the travel time grids')
        t0 = time()
        seeds = self.srces.locs
        seed_labels = self.srces.labels

        tt = self.velocities.to_time(seeds, seed_labels,
                                     multi_threaded=multi_threaded)

        self.travel_times = tt

        # if self.has_p_velocity():
        #     tt_gs_p = self.p_velocity.to_time(seeds, seed_labels,
        #                                       multi_threaded=multi_threaded)
        #     self.travel_times = tt_gs_p
        # if self.has_s_velocity():
        #     tt_gs_s = self.s_velocity.to_time(seeds, seed_labels,
        #                                       multi_threaded=multi_threaded)
        #     self.travel_times = tt_gs_s
        #
        # if self.p_velocity and self.s_velocity:
        #     self.travel_times = tt_gs_p + tt_gs_s

        # cleaning the directory before writing the new files

        for fle in self.paths.times.glob('*time*'):
            fle.unlink(missing_ok=True)

        self.travel_times.write(self.paths.times)
        # self.travel_times.write_hdf5(self.paths. / 'test.hf5')
        t1 = time()
        logger.info(f'done initializing the travel time grids in '
                    f'{t1 - t0:0.2f} seconds')

    def add_inventory(self, inventory, create_srces_file: bool = True,
                      initialize_travel_time: bool = True):
        """
        adding a inventory object to the project
        :param inventory: station xml inventory object
        :type inventory: uquake.core.inventory.Inventory
        :param create_srces_file: if True create or replace the srces file
        (Default: True)
        :param initialize_travel_time: if True the travel time grids are
        initialized (default: True)
        :return:
        """

        inventory.write(str(self.files.inventory))
        self.srces = Srces.from_inventory(inventory)
        if create_srces_file:
            with open(self.files.srces, 'wb') as srces_file:
                pickle.dump(self.srces, srces_file)

        if initialize_travel_time:
            self.init_travel_time_grids()
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

    def add_inventory_from_file(self, inventory_file: str, format='StationXML',
                                create_srces_file: bool = True,
                                initialize_travel_time: bool = True,
                                xy_from_lat_lon: bool = True,
                                input_projection=4326,
                                output_projection=None):
        """
        adding a inventory object to the project from a file
        :param inventory_file: station xml inventory object
        :type inventory_file: uquake.core.inventory.Inventory
        :param format: file format for the inventory file
        :param create_srces_file: if True create or replace the srces file
        (Default: True)
        :param initialize_travel_time: if True the travel time grids are
        initialized (default: True)
        :param xy_from_lat_lon: convert latitude and longitude in the file to
        cartesian coordinates
        :param input_projection: input projection (default=4326 for typical
        latitude and longitude)
        :param output_projection: output projection (default None)
        :return:
        """

        inventory = read_inventory(inventory_file, format=format,
                                   xy_from_lat_lon=xy_from_lat_lon,
                                   input_projection=input_projection,
                                   output_projection=output_projection)

        self.add_inventory(inventory, create_srces_file=create_srces_file,
                           initialize_travel_time=initialize_travel_time)

    def add_srces(self, srces, force=False, initialize_travel_time: bool=True,
                  multi_threaded: bool=True):
        """
        add a list of sources to the projects
        :param srces: list of sources or sites
        :param force: force the insertion of the srces object if an inventory
        file is present
        :param initialize_travel_time: if True, initialize the travel time
        grid
        :type srces: Srces
        :param multi_threaded: if true, use multi-threading
        :type multi_threaded: bool

        ..warning:: travel time should be initialized when the sites/srces
        are updated. Not doing so, may cause the sites/source and the
        travel time grids to be incompatible.
        """

        if not isinstance(srces, Srces):
            raise TypeError(f'Expecting type {Srces}, given '
                            f'{type(srces)}')

        if self.inventory is not None:
            logger.warning('The project already has an inventory file!')
            if not force:
                logger.warning('exiting...')
                return
            else:
                logger.warning('the force flag value is True, srces object '
                               'will be added.')

        self.srces = srces
        with open(self.files.srces, 'wb') as srces_file:
            pickle.dump(self.srces, srces_file)

        if initialize_travel_time:
            self.init_travel_time_grids(multi_threaded=multi_threaded)
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

    def add_velocities(self, velocities, initialize_travel_times=False):
        """
        add P- and S-wave velocity models to the project
        :param velocities: velocity models
        :type velocities: nlloc.grid.VelocityEnsemble
        :param initialize_travel_times: if True, initialize the travel time
        grid

        ..warning:: travel time should be initialized when the sites/srces
        are updated. Not doing so, may cause the sites/source and the
        travel time grids to be incompatible.
        """

        # velocities.write(path=self.paths.velocities)

        self.velocities = velocities

        for key in velocities.keys():
            self.add_velocity(velocities[key])

        if initialize_travel_times:
            self.init_travel_time_grids()

    def velocity_versioning_handler(self, phase):
        """
        returns the version number of the to be archived velocity model
        :return: int
        """
        file_list = [fle for fle in self.paths.archives.glob(f'*.{phase}.*')]
        # file_list = os.listdir(self.paths.archives / )
        if not file_list:
            return 1

        exts = []
        for element in file_list:
            ext = int(element.split('_')[-1])
            exts.append(ext)

        return np.max(exts) + 1

    def add_velocity(self, velocity, initialize_travel_times=False):
        """
        add P- or S-wave velocity model to the project
        :param velocity: p-wave velocity model
        :type velocity: uquake.grid.nlloc.VelocityGrid3D
        :param initialize_travel_times: if true initialize the travel time grids

        ..warning:: travel time should be initialized when the sites/srces
        are updated. Not doing so, may cause the sites/source and the
        travel time grids to be incompatible.
        """

        if velocity.phase.upper() == 'P':
            self.p_velocity = velocity
        elif velocity.phase.upper() == 'S':
            self.s_velocity = velocity

        velocity.write(self.paths.velocities)

        if initialize_travel_times:
            try:
                self.init_travel_time_grids()
            except Exception as e:
                logger.warning(e)

    def list_active_projects(self):
        """
        List all active projects
        :return: list of active projects
        """
        projects = []
        for project in self.base_projects_path.glob('*'):
            if project.name == 'archives':
                continue
            if project.is_dir():
                projects.append(project.name)

        return projects

    def list_active_networks_project(self, project=None):
        """
        List all active networks
        :param project: the project for which to list the networks
        :return: a list of networks
        """
        if project is None:
            project = self.project_name

        project_directory = self.base_projects_path / project

        if not project_directory.exists():
            logger.warning(f'the project {project} does not exists')
            return

        networks = []
        for network in project_directory.glob('*'):
            if not network.is_dir():
                continue
            networks.append(network.name)

        return networks

    def list_all_active_networks(self):
        """
        list all active projects and networks
        :return: a dictionary of list
        """

        out_dict = {}
        for project in self.list_active_projects():
            networks = self.list_active_networks_project(project=project)
            out_dict[project] = networks

        return out_dict

    def list_all_archived_networks(self):
        """
        List all archived projects and networks
        :return: dictionary of list where the keys are the projects name
        """

        out_dict = {}
        for project in self.paths.archives.glob('*'):
            if not project.is_dir():
                continue
            archived_project_dir = self.paths.archives / project.name
            out_dict[project.name] = []
            for network in archived_project_dir.glob('*'):
                if not network.is_dir():
                    continue
                out_dict[project.name].append(network.name)

        return out_dict

    def has_p_velocity(self):
        if self.p_velocity:
            return True
        return False

    def has_s_velocity(self):
        if self.s_velocity:
            return True
        return False

    def has_inventory(self):
        if self.inventory or self.srces:
            return True
        return False





from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy import Column, Integer, String, DateTime
from uquake.core.stream import Trace

Base = declarative_base()


class Database(object):
    def __init__(self, url, port, username, password, db):
        self.connection_string=f'postgresql://{username}:{password}' \
                               f'@{url}:{port}/db'
        self.engine = create_engine(self.connection_string)
        Base.metadata.create_all(self.engine)
        self.db_session = sessionmaker(bind=self.engine)
        self.session = self.db_session()

    def index_trace(self, tr: Trace, file_path):
        csd = ContinuousSeismicData(network=tr.stats.network,
                                    station=tr.stats.station,
                                    location=tr.stats.location,
                                    channel=tr.stats.channel,
                                    start_time=tr.stats.starttime.datetime,
                                    end_time=tr.stats.endtime.datetime,
                                    filepath=file_path)
        self.session.add(csd)
        self.session.commit()


class ContinuousSeismicData(Base):
    __tablename__ = "continuous_seismic_data"
    id = Column(Integer, primary_key=True, index=True)
    network = Column(String, index=True)
    station = Column(String, index=True)
    location = Column(String, index=True)
    channel = Column(String, index=True)
    start_time = Column(DateTime, index=True)
    end_time = Column(DateTime, index=True)
    filepath = Column(String)





from torch.utils.data import Dataset
import torch
import os
from PIL import Image
import pandas as pd
import numpy as np
from glob import glob
from pathlib import Path
from os.path import sep
from loguru import logger
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
from uquake.core.trace import Trace
from abc import ABC
import pickle


# class Label():
#     def __init__(self, labels):
#         self.full_labels = list(labels)
#         self.labels = range(0, len(labels))
#
#         self.unique_labels = np.unique(labels)
#         self.unique_numerical_labels = np.range(len(self.unique_labels))
#
#         self.labels_dict = {}
#         for full_label, label  in zip(labels, self.unique_labels):
#             self.labels_dict[full_label] = label
#
#     def __len__(self):
#         return len(self.labels)
#
#     def __getitem__(self, item):
#         label = self.full_labels[item]
#         return self.labels_dict[label]
#
#     def get_full_label(self, item):

def get_file_list(input_directory, suffix, extension='png'):
    path = os.path.join(input_directory, '*', f'*{suffix}.{extension}')
    return glob(path)


def split_dataset(path, split=0.8, seed=None, max_item_per_category=100000):
    """

    :param path: list of image files
    :param split: set_1 fraction (between 0 and 1)
    :param seed: seed for the random number generator to provide
    reproducibility (Default None)
    :param max_item_per_category: maximum number of item per category
    :return: set_1, set_2
    """

    np.random.seed(seed)
    choices = np.random.choice([False, True], size=len(file_list),
                               p=[1-split, split])
    not_choices = [not choice for choice in choices]
    set_1 = np.array(file_list)[choices]
    set_2 = np.array(file_list)[not_choices]

    return set_1, set_2


def split_filename_label(file_list):
    file_list = [os.path.split(f)[-1] for f in tmp]
    labels = [os.path.split(f)[0].split('/')[-1] for f in tmp]


class FileList(object):
    def __init__(self, path, unused_category=['unknown'], seed=None,
                 extension='.jpg'):

        self.path = Path(path)
        dir_list = self.path.glob('*')

        self.category_list = [str(dr).split(sep)[-1] for dr in dir_list
                              if str(dr).split(sep)[-1] not in unused_category]
        self.labels = np.arange(len(self.category_list))

        self.labels_dict = {}
        np.random.seed(seed)
        self.cat_file_list = {}
        for label, category in zip(self.labels, self.category_list):
            logger.info(f'processing {category}')
            self.labels_dict[category] = label

            self.cat_file_list[category] = [fle for fle in
                                            (self.path /
                                             category).glob(f'*.{extension}')]

        self.classes = self.category_list

    def select(self, number):
        return SpectrogramDataset.from_file_list(self,
                                                 max_number_image_per_category
                                                 =number)

    def select1d(self, number):
        return ClassifierDataset1D.from_file_list(self,
                max_number_signal_per_category=number)


class SpectrogramDataset(Dataset):

    def __init__(self, file_dict: dict,
                 max_number_image_per_category: int = 1e5,
                 seed: int = None):

        categories = [key for key in file_dict.keys()]
        self.category_list = categories
        self.labels = np.arange(len(categories))

        np.random.seed(seed)
        self.labels_dict = {}
        self.file_list = []
        self.label_list = []
        for category, label in zip(categories, self.labels):
            self.labels_dict[category] = label
            nb_files = int(max_number_image_per_category)
            if len(file_dict[category]) < nb_files:
                nb_files = int(len(file_dict[category]))
            for f in np.random.choice(file_dict[category],
                                      size=nb_files,
                                      replace=False):
                self.file_list.append(f)
                self.label_list.append(label)

        self.file_list = np.array(self.file_list)
        self.label_list = np.array(self.label_list)

    @classmethod
    def from_file_list(cls, file_list: FileList,
                       max_number_image_per_category=1e5, seed=None):
        return cls(file_list.cat_file_list,
                   max_number_image_per_category=max_number_image_per_category,
                   seed=seed)

    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # label = self.df_data.iloc[idx]['label']
        # category = self.category_list[idx]
        label = self.label_list[idx]

        image = torch.from_numpy((np.array(Image.open(
            self.file_list[idx])) / 255).astype(np.float32))

        # return {'data': image, 'label': label}
        return image, label

    @property
    def shape(self):
        px_x, px_y = self[0][0].shape
        return len(self.label_list), px_x, px_y

    @property
    def nb_pixel(self):
        return np.prod(self[0][0].shape)

    @property
    def categories(self):
        return self.labels_dict

    @property
    def classes(self):
        return self.categories

    @property
    def nb_categories(self):
        return len(self.category_list)


sampling_rate = 6000
# num_threads = int(np.ceil(cpu_count() - 10))
num_threads = 10
replication_level = 5
snr_threshold = 10
sequence_length_second = 2
perturbation_range_second = 1
image_width = 128
image_height = 128
buffer_image_fraction = 0.05

buffer_image_sample = int(image_width * buffer_image_fraction)

hop_length = int(sequence_length_second * sampling_rate //
                 (image_width + 2 * buffer_image_sample))


def spectrogram(trace: Trace):

    trace.resample(sampling_rate)

    mel_spec = MelSpectrogram(sample_rate=sampling_rate,
                              n_mels=image_height,
                              hop_length=hop_length,
                              power=1,
                              pad_mode='reflect',
                              normalized=True)

    amplitude_to_db = AmplitudeToDB()

    # trace = trace.detrend('linear')
    # trace = trace.detrend('demean')
    trace.data = trace.data - np.mean(trace.data)
    trace = trace.taper(max_length=0.01, max_percentage=0.05)
    trace = trace.trim(starttime=trace.stats.starttime,
                       endtime=trace.stats.starttime + sequence_length_second,
                       pad=True, fill_value=0)
    data = trace.data

    torch_data = torch.tensor(data).type(torch.float32)

    spec = (mel_spec(torch_data))
    spec_db = amplitude_to_db(spec.abs() + 1e-3)
    spec_db = (spec_db - spec_db.min()).numpy()
    # spec_db = (spec_db / spec_db.max()).type(torch.float32)
    return spec_db


class ClassifierDataset1D(Dataset):

    def __init__(self, file_dict: dict,
                 max_number_signal_per_category: int = 1e5,
                 seed: int = None):

        categories = [key for key in file_dict.keys()]
        self.category_list = categories
        self.labels = np.arange(len(categories))

        np.random.seed(seed)
        self.labels_dict = {}
        self.file_list = []
        self.label_list = []
        for category, label in zip(categories, self.labels):
            self.labels_dict[category] = label
            nb_files = int(max_number_signal_per_category)
            if len(file_dict[category]) < nb_files:
                nb_files = int(len(file_dict[category]))
            for f in np.random.choice(file_dict[category],
                                      size=nb_files,
                                      replace=False):
                self.file_list.append(f)
                self.label_list.append(label)

        self.file_list = np.array(self.file_list)
        self.label_list = np.array(self.label_list)

    @classmethod
    def from_file_list(cls, file_list: FileList,
                       max_number_signal_per_category=1e5, seed=None):
        return cls(file_list.cat_file_list,
                   max_number_signal_per_category=
                   max_number_signal_per_category,
                   seed=seed)

    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        with open(self.file_list[idx], 'rb') as f_in:
            data_dict = pickle.load(f_in)

        # label = self.df_data.iloc[idx]['label']
        # category = self.category_list[idx]
        label = self.label_list[idx]

        data = data_dict['data']

        if idx != 0:
            data_out = np.zeros(self.shape[1])
            sample_diff = len(data) - self.shape[1]
            if sample_diff > 0:
                data_out = data[0:self.shape[1]]
            else:
                data_out[0: len(data)] = data
        else:
            data_out = data

        data_out -= np.mean(data_out)
        data_out /= np.max(data_out)

        data_out = torch.from_numpy((np.array(data_out).astype(np.float32)))

        return data_out, label

    @property
    def shape(self):
        len_data = len(self[0][0])
        return len(self.label_list), len_data

    @property
    def categories(self):
        return self.labels_dict

    @property
    def classes(self):
        return self.categories

    @property
    def nb_categories(self):
        return len(self.category_list)


class PickingDataset(Dataset):
    def __init__(self, file_list):
        self.file_list = file_list

    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        with open(self.file_list[idx], 'rb') as f_in:
            data = pickle.load(f_in)

        signal = data['data'].astype(np.float32)

        # ensuring the signal is between -1 and 1
        signal = signal - np.mean(signal)
        signal = signal / np.max(signal)

        target = data['pick']

        return signal.astype(np.float32), np.float32(target)

    @property
    def shape(self):
        len_signal = self[0][0].shape
        return 1, len_signal[0]

    def split(self, split_fraction: float = 0.8, random_seed: int = None):
        np.random.seed(random_seed)
        choices = np.random.choice([False, True], size=len(self),
                                   p=[1 - split_fraction, split_fraction])
        not_choices = np.invert(choices)
        files_set_a = np.array(self.file_list)[choices]
        files_set_b = np.array(self.file_list)[not_choices]

        set_a = PickingDataset(files_set_a)
        set_b = PickingDataset(files_set_b)

        return set_a, set_b










from os import environ, walk
from glob import glob
from uquake.core import read
from .database import Database

data_path = environ['UQ_FS_DATA_PATH']
database = environ['UQ_FS_DATABASE']
db_user = environ['UQ_FS_DB_USER']
db_password = environ['UQ_FS_DB_PASSWORD']
db_url = environ['UQ_FS_DB_URL']
db_port = environ['UQ_FS_DB_PORT']

print(len(glob(data_path + '/*', recursive=True)))

db = Database(db_url, db_port, db_user, db_password, database)

for (dirpath, dirnames, filenames) in walk(data_path):
    for filename in filenames:
        st = read(dirpath + '/' + filename)

        for tr in st:

# for file in glob(data_path, recursive=True):






from .event import SimpleArrival
from pydantic import BaseModel
from typing import Optional, List
from uquake.nlloc import nlloc
from enum import Enum
from datetime import datetime
import numpy as np
from uquake.core import UTCDateTime
from .base import Coordinates3D
from .event import Ray
import useis
from uquake.core.event import Event


class FloatType(str, Enum):
    FLOAT = "FLOAT"
    DOUBLE = "DOUBLE"


class Site(BaseModel):
    label: str
    location: Coordinates3D


class Srces(BaseModel):
    sites: List[Site]


class Observations(BaseModel):
    picks: List[SimpleArrival]
    p_pick_error: Optional[float] = 1e-3
    s_pick_error: Optional[float] = 1e-3

    class Config:
        orm_mode = True

    @classmethod
    def from_uquake(cls, observations: nlloc.Observations):
        simple_arrivals = []
        for pick in observations.picks:
            simple_arrivals.append(SimpleArrival.from_uquake_pick(pick))

        return cls(picks=simple_arrivals,
                   p_pick_error=observations.p_pick_error,
                   s_pick_error=observations.s_pick_error)

    def to_uquake(self):
        picks = []
        for pick in self.picks:
            picks.append(pick.to_uquake_pick())
        return nlloc.Observations(picks=picks,
                                  p_pick_error=self.p_pick_error,
                                  s_pick_error=self.s_pick_error)

    def to_dict(self):
        picks = []
        for pick in self.picks:
            picks.append(pick.__dict__)

        observations_dict = self.__dict__
        observations_dict['picks'] = picks
        return observations_dict


class NLLOCResults(BaseModel):
    hypocenter: Coordinates3D
    event_time: datetime
    scatter_cloud: List[Coordinates3D]
    rays: List[Ray]
    observations: Observations
    uncertainty: float
    hypocenter_file: str

    @classmethod
    def from_nlloc_results(cls, nlloc_results: useis.procesors.NLLocResults):
        hypocenter = Coordinates3D.from_array(nlloc_results.hypocenter)

        scatter_cloud = []
        for scatter in nlloc_results.scatter_cloud:
            scatter_cloud.append(Coordinates3D.from_array(scatter))

        rays = []
        for ray in nlloc_results.rays:
            rays.append(Ray.from_uquake(ray))

        observations = Observations.from_uquake(nlloc_results.observations)

        return cls(hypocenter=hypocenter, event_time=nlloc_results.event_time,
                   scatter_cloud=scatter_cloud, rays=rays,
                   observations=observations,
                   uncertainty=nlloc_results.uncertainty,
                   hypocenter_file=nlloc_results.hypocenter_file)



from torchvision import models
from .dataset import (Dataset, PickingDataset, ClassifierDataset1D,
                      spectrogram)
from torch.utils.data import DataLoader
import torch
from torch import nn
import numpy as np
from loguru import logger
from tqdm import tqdm
import pickle
from sklearn.metrics import confusion_matrix
# from plotcm import plot_confusion_matrix
from uquake.core import Stream, Trace
# from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
from PIL import Image, ImageOps
from torch import nn
from .resnet1d import ResNet1D
from importlib import reload
# from .dataset as dataset
# reload(dataset)


sampling_rate = 6000
# num_threads = int(np.ceil(cpu_count() - 10))
num_threads = 10
replication_level = 5
snr_threshold = 10
sequence_length_second = 2
perturbation_range_second = 1
image_width = 128
image_height = 128
buffer_image_fraction = 0.05


class EventClassifier(object):
    def __init__(self, n_features: int, gpu: bool = True):

        # define the model
        self.model = models.resnet34(pretrained=False)
        self.model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3,
                                     bias=False)

        self.n_features = n_features

        self.model.fc = nn.Linear(self.model.fc.in_features, n_features)

        self.optimizer = torch.optim.Adam(self.model.parameters())
        self.criterion = nn.CrossEntropyLoss()
        self.gpu = gpu

        if gpu:
            device = torch.device("cuda:0" if
                                  torch.cuda.is_available() else "cpu")
            if device == "cpu":
                logger.warning('GPU is not available, the CPU will be used')
            else:
                logger.info('GPU will be used')
        else:
            device = 'cpu'
            logger.info('The CPU will be used')

        self.device = device

        self.model = self.model.to(self.device)
        self.display_memory()
        self.accuracies = []
        self.losses = []

        self.validation_predictions = None
        self.validation_targets = None

    @classmethod
    def from_pretrained_model(cls, model, gpu: bool = True):
        n_features = model.fc.out_features
        cls_tmp = cls(n_features, gpu=gpu)
        cls_tmp.model = model.to(cls_tmp.device)

        return cls_tmp

    @classmethod
    def from_pretrained_model_file(cls, path, gpu: bool = True):
        with open(path, 'rb') as input_file:
            model = pickle.load(input_file)

        return cls.from_pretrained_model(model, gpu=gpu)

    # @classmethod
    # def load_model(cls, file_name):
    #     return pickle.load(open('filename', 'rb'))

    def train(self, dataset: Dataset, batch_size: int):

        self.model.train()

        self.accuracies = []
        self.losses = []

        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1],
                                 inputs.size()[2])
            self.iterate(inputs, targets)
            torch.cuda.empty_cache()

    def iterate(self, inputs, targets):

        self.optimizer.zero_grad()

        inputs = inputs.to(self.device)
        targets = targets.to(self.device)

        # with torch.no_grad():
        predictions = self.model(inputs)

        loss = self.criterion(predictions, targets)
        # loss.requires_grad = True
        loss.backward()
        self.optimizer.step()

        self.accuracies.append(self.measure_accuracy(targets,
                                                     predictions).item())
        self.losses.append(loss.item())

    def validate(self, dataset: Dataset, batch_size: int):
        self.model.eval()
        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)
        self.validation_targets = np.zeros(len(dataset))
        self.validation_predictions = np.zeros(len(dataset))
        accuracy = 0
        for i, (inputs, targets) in enumerate(tqdm(training_loader)):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1],
                                 inputs.size()[2])
            inputs = inputs.to(self.device)

            start = batch_size * i
            end = start + batch_size
            self.validation_targets[start:end] = targets
            targets = targets.to(self.device)

            # with torch.no_grad() pytorch does not compute the gradient,
            # the gradient takes a significant amount of memory on the gpu.
            # using torch.no_grad() consequently allows significantly
            # increasing the batch size
            with torch.no_grad():
                outputs = self.model(inputs)

            self.validation_predictions[start:end] = outputs.max(dim=1)[1].cpu(
            ).detach().numpy()

            accuracy += self.measure_accuracy(targets, outputs)
        self.model.train()
        return accuracy.cpu().detach().numpy() / (i + 1)

    def display_memory(self):
        torch.cuda.empty_cache()
        memory = torch.cuda.memory_allocated(self.device)
        logger.info("{:.3f} GB".format(memory / 1024 ** 3))

    def save(self, file_name):
        return pickle.dump(self, open(file_name, 'wb'))

    def write(self, file_name):
        return self.save(file_name)

    @classmethod
    def read(cls, file_name):
        with open(file_name, 'rb') as f_in:
            return pickle.load(f_in)

    @staticmethod
    def measure_accuracy(targets, predictions):
        max_index = predictions.max(dim=1)[1]
        return (max_index == targets).sum() / len(targets)

    def predict(self, stream: Stream):
        """

        :param stream: A uquake.core.stream.Stream object containing the
        waveforms
        :return:
        """
        specs = generate_spectrogram(stream)
        return self.predict_spectrogram(specs)

    def predict_spectrogram(self, specs):
        self.model.eval()
        if isinstance(specs, list):
            specs = np.array(specs)
        if len(specs.shape) == 2:
            specs = torch.from_numpy(specs).view(1, 1, specs.shape[0],
                                                 specs.shape[1])
        else:
            specs = torch.from_numpy(specs).view(specs.shape[0], 1,
                                                 specs.shape[1],
                                                 specs.shape[2])

        specs = specs.to(self.device)
        # self.model.train()
        with torch.no_grad():
            predictions = (self.model(specs).argmax(axis=1).cpu(),
                           self.model(specs).cpu())

        self.model.train()
        return predictions


def read_classifier(file_name):
    return EventClassifier.read(file_name)


def generate_spectrogram(stream: Stream):

    specs = []
    trs = []
    for tr in stream.copy():
        # check if tr contains NaN
        if np.any(np.isnan(tr.data)):
            continue
        trs.append(tr.copy())

    st2 = Stream(traces=trs)
    st2 = st2.trim(endtime=tr.stats.starttime + 2, pad=True,
                 fill_value=0)
    st2 = st2.detrend('demean').detrend('linear')

    st2 = st2.taper(max_percentage=1, max_length=1e-2)

    for tr in st2:
        spec = spectrogram(tr)
        spec /= np.max(spec)

        specs.append(spec)

    return specs


class EventClassifier1D(EventClassifier):
    def __init__(self, n_classes: int, in_channels: int = 1,
                 base_filters: int = 16, kernel_size: int = 7,
                 stride: int = 3, groups: int = 1, n_block: int = 8,
                 learning_rate=1e-3, gpu: bool = True):

        if gpu:
            device = torch.device("cuda:0" if
                                  torch.cuda.is_available() else "cpu")
            if device == "cpu":
                logger.warning('GPU is not available, the CPU will be used')
            else:
                logger.info('GPU will be used')
        else:
            device = 'cpu'
            logger.info('The CPU will be used')

        self.device = device

        self.model = ResNet1D(in_channels, base_filters, kernel_size,
                              stride, groups, n_block,
                              n_classes).to(self.device)

        # self.model = Conv1DModel()

        self.in_channels = in_channels
        self.base_filters = base_filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.n_classes = n_classes
        self.groups = groups
        self.n_block = n_block
        self.gpu = gpu

        self.model = ResNet1D(in_channels, base_filters, kernel_size,
                              stride, groups, n_block,
                              n_classes).to(self.device)

        self.optimizer = torch.optim.Adam(self.model.parameters(),
                                          lr=learning_rate)
        # self.optimizer = torch.optim.SGD(self.model.parameters(),
        #                                  lr=learning_rate)
        self.criterion = nn.CrossEntropyLoss()

        self.display_memory()
        self.losses = []

        self.validation_predictions = None
        self.validation_targets = None

    def train(self, dataset: ClassifierDataset1D, batch_size: int):

        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])
            self.iterate(inputs, targets)
            torch.cuda.empty_cache()

    def iterate(self, inputs, targets):

        self.optimizer.zero_grad()

        inputs = inputs.to(self.device)
        targets = targets.to(self.device)

        predictions = self.model(inputs)

        # loss = self.criterion(predictions, targets.view(len(targets), -1))
        loss = self.criterion(predictions, targets)
        # print(loss.item())
        loss.backward()
        self.optimizer.step()

        # print(loss.item())
        self.losses.append(loss.cpu().item())

    def validate(self, dataset: ClassifierDataset1D, batch_size: int):
        self.model.eval()
        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)
        self.validation_targets = np.zeros(len(dataset))
        self.validation_predictions = np.zeros(len(dataset))
        accuracy = 0
        for i, (inputs, targets) in enumerate(tqdm(training_loader)):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])
            inputs = inputs.to(self.device)

            start = batch_size * i
            end = start + batch_size
            self.validation_targets[start:end] = targets
            targets = targets.to(self.device)

            # with torch.no_grad() pytorch does not compute the gradient,
            # the gradient takes a significant amount of memory on the gpu.
            # using torch.no_grad() consequently allows significantly
            # increasing the batch size
            with torch.no_grad():
                outputs = self.model(inputs)

            self.validation_predictions[start:end] = outputs.max(dim=1)[1].cpu(
            ).detach().numpy()

            accuracy += self.measure_accuracy(targets, outputs)
        self.model.train()
        return accuracy.cpu().detach().numpy() / (i + 1)


class Conv1DModel(nn.Module):
    def __init__(self):
        super(Conv1DModel, self).__init__()
        self.conv_layers = nn.Sequential(
            nn.Conv1d(1, 64, 15),
            nn.ReLU(),
            nn.Conv1d(64, 64, 15),
            nn.ReLU(),
            nn.Conv1d(64, 64, 15),
            nn.ReLU(),
            nn.Conv1d(64, 128, 15),
            nn.ReLU(),
            nn.Conv1d(128, 128, 15),
            nn.ReLU(),
            nn.Conv1d(128, 128, 15),
            nn.ReLU()
        )

        self.dense_layers = nn.Sequential(
            nn.Dropout(0.2),
            nn.Linear(in_features=128 * 172, out_features=256),
            nn.ReLU(),
            nn.Linear(in_features=256, out_features=256),
            nn.ReLU(),
            nn.Linear(in_features=256, out_features=256),
            nn.ReLU(),
            nn.Linear(in_features=256, out_features=1))

    def forward(self, X):
        out = self.conv_layers(X)
        out = out.view(out.size(0), -1)
        out = self.dense_layers(out)
        # input(out)
        return out


class AIPicker(EventClassifier):
    def __init__(self, in_channels: int = 1, base_filters: int = 64,
                 kernel_size: int = 15, stride: int = 3, n_classes: int = 1,
                 groups: int = 1, n_block: int = 16, learning_rate=1e-3,
                 gpu: bool = True):
        self.in_channels = in_channels
        self.base_filters = base_filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.n_classes = n_classes
        self.groups = groups
        self.n_block = n_block
        self.gpu = gpu

        if gpu:
            device = torch.device("cuda:0" if
                                  torch.cuda.is_available() else "cpu")
            if device == "cpu":
                logger.warning('GPU is not available, the CPU will be used')
            else:
                logger.info('GPU will be used')
        else:
            device = 'cpu'
            logger.info('The CPU will be used')

        self.device = device

        self.model = ResNet1D(in_channels, base_filters, kernel_size,
                              stride, groups, n_block,
                              n_classes).to(self.device)

        # self.model = PickerModel().to(device)

        # self.model = nn.Sequential(
        #     nn.Conv1d(1, 64, 16),
        #     nn.Conv1d(64, 64, 16),
        #     nn.Conv1d(64, 64, 16),
        #     nn.Conv1d(64, 128, 16),
        #     nn.Conv1d(128, 128, 16),
        #     nn.Conv1d(128, 128, 16),
        #     nn.Linear(in_features=12 * 15, out_features=120),
        #     nn.Linear(in_features=120, out_features=60),
        #     nn.Linear(in_features=60, out_features=30),
        #     nn.Linear(in_features=30, out_features=1)
        # ).to(device)

        self.optimizer = torch.optim.Adam(self.model.parameters(),
                                          lr=learning_rate)
        # self.optimizer = torch.optim.SGD(self.model.parameters(),
        #                                  lr=learning_rate)
        self.criterion = nn.MSELoss()

        self.display_memory()
        self.losses = []

        self.validation_predictions = None
        self.validation_targets = None

    def train(self, dataset: PickingDataset, batch_size: int):

        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])
            self.iterate(inputs, targets)
            torch.cuda.empty_cache()

    def iterate(self, inputs, targets):

        self.optimizer.zero_grad()

        inputs = inputs.to(self.device)
        targets = targets.to(self.device)

        predictions = self.model(inputs)

        loss = self.criterion(predictions, targets.view(len(targets), -1))
        # print(loss.item())
        loss.backward()
        self.optimizer.step()

        # print(loss.item())
        self.losses.append(loss.cpu().item())

    def predict(self, inputs):
        if isinstance(inputs, np.ndarray):
            if len(inputs.shape) == 1:
                dim1 = 1
                dim2 = inputs.shape[0]
            else:
                dim1 = inputs.shape[0]
                dim2 = inputs.shape[1]

            inputs = torch.from_numpy(inputs)
        elif not isinstance(inputs, torch.Tensor):
            raise TypeError

        inputs = inputs.view(dim1, -1, dim2).to(self.device)
        with torch.no_grad():
            outputs_tensor = self.model(inputs)

        outputs = [out.item() for out in outputs_tensor.cpu()]
        return outputs

    def validate(self, dataset: PickingDataset, batch_size):
        predictions = []
        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        losses = []
        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            with torch.no_grad():
                ps = self.model(inputs)

            for p in ps.cpu():
                predictions.append(p.item())

            loss = self.criterion(ps, targets.view(len(targets), -1))
            # print(loss.cpu().item())
            losses.append(loss.cpu().item())
        return predictions, np.mean(losses), np.std(losses)

    def save(self, file_name):
        with(open(file_name, 'wb')) as f_out:
            pickle.dump(self, f_out)

    @classmethod
    def read(cls, file_name):
        with(open(file_name, 'rb')) as f_in:
            return pickle.load(f_in)


def read_picker(file_name):
    return AIPicker.read(file_name)






import requests
import pickle
from uquake.grid.nlloc import VelocityGrid3D
from uquake.core.inventory import Inventory
from uquake.nlloc.nlloc import Srces, Observations
from io import BytesIO
from furl import furl
from typing import Optional, List
import json
import numpy as np
from useis.services.models.grid import VelocityGrid3D as ModelVelocityGrid3D
from useis.services.models.nlloc import Observations as ModelObservations
import json


def serialize_object(obj, name):
    f_out = BytesIO()
    pickle.dump(obj, f_out)
    f_out.seek(0)
    f_out.name = name
    return f_out


class GridClient:
    def __init__(self, base_url, project, network, user=None, password=None):
        self.base_url = furl(base_url)
        self.project = project
        self.network = network
        self.user = user
        self.password = password

    def add_3d_velocity(self, velocity: VelocityGrid3D,
                        initialize_travel_times=False):
        url = self.base_url / 'velocity' / self.project / self.network \
              / 'add' / '3D'

        data = {'initialize_travel_times': initialize_travel_times}

        velocity_grid = ModelVelocityGrid3D.from_uquake(velocity)
        params = {'velocity': velocity_grid,
                  'initialize_travel_times': initialize_travel_times}

        return requests.post(url, params=data, json=velocity_grid.to_dict())

    def add_inventory(self, inventory: Inventory,
                      initialize_travel_times=False):
        f_out = BytesIO()
        inventory.write(f_out)
        f_out.seek(0)
        f_out.name = 'inventory'
        url = self.base_url / 'inventory' / self.project / self.network
        files = {'inventory': f_out}
        data = {'initialize_travel_times': initialize_travel_times}
        return requests.post(url, files=files, data=data)

    def add_srces(self, srces: Srces, initialize_travel_times=False):
        f_out = serialize_object(srces, 'srces')
        url = self.base_url / 'srces' / self.project / self.network
        files = {'srces': f_out}
        data = {'initialize_travel_times': initialize_travel_times}
        return requests.post(url, files=files, data=data)

    def add_srces(self, srces: Srces, initialize_travel_times=False):
        f_out = serialize_object(srces, 'srces')
        url = self.base_url / 'srces' / self.project / self.network
        files = {'srces': f_out}
        data = {'initialize_travel_times': initialize_travel_times}
        return requests.post(url, files=files, data=data)

    def initialize_travel_times(self, multi_threaded=False):
        data = {'multi_threaded': multi_threaded}
        url = self.base_url / 'travel_times' / self.project / self.network / \
              'init'
        return requests.get(url, data=data)

    def event_location(self, observations: Observations):
        url = self.base_url / 'event' / 'locate' / self.project / self.network
        model_observations = ModelObservations.from_uquake(observations)
        from ipdb import set_trace
        set_trace()
        return requests.post(url, json=model_observations.to_dict())

    def generate_random_points_in_grid(self, n_points: Optional[int] = 1):
        url = self.base_url / 'test' / self.project / self.network / \
              'random_locations'

        params = {'n_points': n_points}
        response = requests.get(url, params=params)
        flat_array = np.array(eval(json.loads(response.content)))
        n2 = int(len(flat_array) / n_points)
        flat_array = flat_array.reshape((n_points, n2))
        return response, flat_array

    def generate_random_observations(self, x, y, z):
        url = self.base_url / 'test' / self.project / self.network / \
            'observations'
        params = {'x': x, 'y': y, 'z': z}
        response = requests.get(url, params=params)
        model_observations = ModelObservations.parse_obj(
            json.loads(response.content))

        return response, model_observations.to_uquake()

    def list_3d_velocity(self):
        url = self.base_url / 'velocity' / self.project / self.network
        return requests.get(url)



"""
resnet for 1-d signal data, pytorch version

Shenda Hong, Oct 2019
"""

import numpy as np
from collections import Counter
from tqdm import tqdm
from matplotlib import pyplot as plt
from sklearn.metrics import classification_report

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader


class MyDataset(Dataset):
    def __init__(self, data, label):
        self.data = data
        self.label = label

    def __getitem__(self, index):
        return (torch.tensor(self.data[index], dtype=torch.float),
                torch.tensor(self.label[index], dtype=torch.long))

    def __len__(self):
        return len(self.data)


class MyConv1dPadSame(nn.Module):
    """
    extend nn.Conv1d to support SAME padding
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride,
                 groups=1):
        super(MyConv1dPadSame, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.groups = groups
        self.conv = torch.nn.Conv1d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            groups=self.groups)

    def forward(self, x):
        net = x

        # compute pad shape
        in_dim = net.shape[-1]
        out_dim = (in_dim + self.stride - 1) // self.stride
        p = max(0, (out_dim - 1) * self.stride + self.kernel_size - in_dim)
        pad_left = p // 2
        pad_right = p - pad_left
        net = F.pad(net, (pad_left, pad_right), "constant", 0)

        net = self.conv(net)

        return net


class MyMaxPool1dPadSame(nn.Module):
    """
    extend nn.MaxPool1d to support SAME padding
    """

    def __init__(self, kernel_size):
        super(MyMaxPool1dPadSame, self).__init__()
        self.kernel_size = kernel_size
        self.stride = 1
        self.max_pool = torch.nn.MaxPool1d(kernel_size=self.kernel_size)

    def forward(self, x):
        net = x

        # compute pad shape
        in_dim = net.shape[-1]
        out_dim = (in_dim + self.stride - 1) // self.stride
        p = max(0, (out_dim - 1) * self.stride + self.kernel_size - in_dim)
        pad_left = p // 2
        pad_right = p - pad_left
        net = F.pad(net, (pad_left, pad_right), "constant", 0)

        net = self.max_pool(net)

        return net


class BasicBlock(nn.Module):
    """
    ResNet Basic Block
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride, groups,
                 downsample, use_bn, use_do, is_first_block=False):
        super(BasicBlock, self).__init__()

        self.in_channels = in_channels
        self.kernel_size = kernel_size
        self.out_channels = out_channels
        self.stride = stride
        self.groups = groups
        self.downsample = downsample
        if self.downsample:
            self.stride = stride
        else:
            self.stride = 1
        self.is_first_block = is_first_block
        self.use_bn = use_bn
        self.use_do = use_do

        # the first conv
        self.bn1 = nn.BatchNorm1d(in_channels)
        self.relu1 = nn.ReLU()
        self.do1 = nn.Dropout(p=0.5)
        self.conv1 = MyConv1dPadSame(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=self.stride,
            groups=self.groups)

        # the second conv
        self.bn2 = nn.BatchNorm1d(out_channels)
        self.relu2 = nn.ReLU()
        self.do2 = nn.Dropout(p=0.5)
        self.conv2 = MyConv1dPadSame(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=1,
            groups=self.groups)

        self.max_pool = MyMaxPool1dPadSame(kernel_size=self.stride)

    def forward(self, x):

        identity = x

        # the first conv
        out = x
        if not self.is_first_block:
            if self.use_bn:
                out = self.bn1(out)
            out = self.relu1(out)
            if self.use_do:
                out = self.do1(out)
        out = self.conv1(out)

        # the second conv
        if self.use_bn:
            out = self.bn2(out)
        out = self.relu2(out)
        if self.use_do:
            out = self.do2(out)
        out = self.conv2(out)

        # if downsample, also downsample identity
        if self.downsample:
            identity = self.max_pool(identity)

        # if expand channel, also pad zeros to identity
        if self.out_channels != self.in_channels:
            identity = identity.transpose(-1, -2)
            ch1 = (self.out_channels - self.in_channels) // 2
            ch2 = self.out_channels - self.in_channels - ch1
            identity = F.pad(identity, (ch1, ch2), "constant", 0)
            identity = identity.transpose(-1, -2)

        # shortcut
        out += identity

        return out


class ResNet1D(nn.Module):
    """

    Input:
        X: (n_samples, n_channel, n_length)
        Y: (n_samples)

    Output:
        out: (n_samples)

    Pararmeters:
        in_channels: dim of input, the same as n_channel
        base_filters: number of filters in the first several Conv layer, it
        will double at every 4 layers
        kernel_size: width of kernel
        stride: stride of kernel moving
        groups: set target to 1 as ResNeXt
        n_block: number of blocks
        n_classes: number of classes

    """

    def __init__(self, in_channels, base_filters, kernel_size, stride, groups,
                 n_block, n_classes, downsample_gap=2, increasefilter_gap=4,
                 use_bn=True, use_do=True, verbose=False):
        super(ResNet1D, self).__init__()

        self.verbose = verbose
        self.n_block = n_block
        self.kernel_size = kernel_size
        self.stride = stride
        self.groups = groups
        self.use_bn = use_bn
        self.use_do = use_do

        self.downsample_gap = downsample_gap  # 2 for base model
        self.increasefilter_gap = increasefilter_gap  # 4 for base model

        # first block
        self.first_block_conv = MyConv1dPadSame(in_channels=in_channels,
                                                out_channels=base_filters,
                                                kernel_size=self.kernel_size,
                                                stride=1)
        self.first_block_bn = nn.BatchNorm1d(base_filters)
        self.first_block_relu = nn.ReLU()
        out_channels = base_filters

        # residual blocks
        self.basicblock_list = nn.ModuleList()
        for i_block in range(self.n_block):
            # is_first_block
            if i_block == 0:
                is_first_block = True
            else:
                is_first_block = False
            # downsample at every self.downsample_gap blocks
            if i_block % self.downsample_gap == 1:
                downsample = True
            else:
                downsample = False
            # in_channels and out_channels
            if is_first_block:
                in_channels = base_filters
                out_channels = in_channels
            else:
                # increase filters at every self.increasefilter_gap blocks
                in_channels = int(base_filters * 2 ** (
                            (i_block - 1) // self.increasefilter_gap))
                if (i_block % self.increasefilter_gap == 0) and (i_block != 0):
                    out_channels = in_channels * 2
                else:
                    out_channels = in_channels

            tmp_block = BasicBlock(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=self.kernel_size,
                stride=self.stride,
                groups=self.groups,
                downsample=downsample,
                use_bn=self.use_bn,
                use_do=self.use_do,
                is_first_block=is_first_block)
            self.basicblock_list.append(tmp_block)

        # final prediction
        self.final_bn = nn.BatchNorm1d(out_channels)
        self.final_relu = nn.ReLU(inplace=True)
        # self.do = nn.Dropout(p=0.5)
        self.dense = nn.Linear(out_channels, n_classes)
        # self.softmax = nn.Softmax(dim=1)

    def forward(self, x):

        out = x

        # first conv
        if self.verbose:
            print('input shape', out.shape)
        out = self.first_block_conv(out)
        if self.verbose:
            print('after first conv', out.shape)
        if self.use_bn:
            out = self.first_block_bn(out)
        out = self.first_block_relu(out)

        # residual blocks, every block has two conv
        for i_block in range(self.n_block):
            net = self.basicblock_list[i_block]
            if self.verbose:
                print('i_block: {0}, in_channels: {1}, '
                      'out_channels: {2}, downsample: {3}'.format(
                        i_block, net.in_channels, net.out_channels,
                        net.downsample))
            out = net(out)
            if self.verbose:
                print(out.shape)

        # final prediction
        if self.use_bn:
            out = self.final_bn(out)
        out = self.final_relu(out)
        out = out.mean(-1)
        if self.verbose:
            print('final pooling', out.shape)
        # out = self.do(out)
        out = self.dense(out)
        if self.verbose:
            print('dense', out.shape)
        # out = self.softmax(out)
        if self.verbose:
            print('softmax', out.shape)

        return out
